use crate::day03::Instr::{Disable, Enable, Mul};

pub fn part1(input: &str) -> u32 {
    parse_instr(input).iter()
        .fold(0u32, |acc, instr| {
            if let Mul(a, b) = instr {
                acc + a * b
            } else {
                acc
            }
        })
}

pub fn part2(input: &str) -> u32 {
    parse_instr(input).iter()
        .scan(true, |enabled, instr| {
            match instr {
                Enable => *enabled = true,
                Disable => *enabled = false,
                Mul(l, r) => return if *enabled { Some(l * r) } else { Some(0) },
            }

            return Some(0);
        }).sum()
}

fn parse_instr(input: &str) -> Vec<Instr> {
    let bytes = input.as_bytes();
    let mut instructions: Vec<Instr> = vec![];

    let mut offset = 0usize;
    while offset < bytes.len() {
        match &bytes[offset..] {
            [b'd', b'o', b'(', b')', ..] => {
                instructions.push(Enable);
                offset += 4;
            }
            [b'd', b'o', b'n', b'\'', b't', b'(', b')', ..] => {
                instructions.push(Disable);
                offset += 7;
            }
            [b'm', b'u', b'l', b'(', rest @ .. ] => {
                let mut stage = 1u8;
                let mut left = 0u32;
                let mut right = 0u32;

                let mut inner_offset = 0usize;
                let mut success = false;
                while inner_offset < rest.len() {
                    match rest[inner_offset] {
                        nb @ b'0'..=b'9' => {
                            if stage == 1 {
                                left = (left * 10) + (nb - b'0') as u32
                            } else {
                                right = (right * 10) + (nb - b'0') as u32
                            }
                        }
                        b',' => {
                            if stage == 1 {
                                stage += 1;
                            } else {
                                break;
                            }
                        }
                        b')' => {
                            if stage == 2 {
                                success = true;
                                instructions.push(Mul(left, right));
                                break;
                            } else {
                                break;
                            }
                        }
                        _ => break
                    }
                    inner_offset += 1;
                }

                if success {
                    offset += 4 + inner_offset;
                } else {
                    offset += 4
                }
            }
            _ => {
                offset += 1
            }
        }
    }

    instructions
}

enum Instr {
    Mul(u32, u32),
    Enable,
    Disable,
}

#[cfg(test)]
mod test {
    use crate::default_tests;
    use super::*;

    const INPUT_1: &str = "xmul(2,4)%&mul[3,7]!@^do_not_mul(5,5)+mul(32,64]then(mul(11,8)mul(8,5))";

    #[test]
    fn test_part1() {
        assert_eq!(part1(INPUT_1), 161);
    }

    default_tests!(day03, 173517243, 100450138);
}
