use crate::day05::ParsingState::{RuleEOL, RuleLeft, RuleRight, Update};
use std::cmp::Ordering;

pub fn part1(input: &str) -> u32 {
    let (rules, updates) = parse(input);

    updates.iter()
        .filter(|update|
            update.windows(2)
                .all(|w| rules[pair_to_idx(&(w[0], w[1]))]))
        .map(|update| update[update.len() / 2] as u32)
        .sum()
}

pub fn part2(input: &str) -> u32 {
    let (rules, updates) = parse(input);

    updates.iter()
        .filter(|update| update.windows(2).any(|w| !rules[pair_to_idx(&(w[0], w[1]))]))
        .map(|update| {
            let mut sorted_update = update.clone();
            sorted_update.sort_by(|first, second| {
                if rules[pair_to_idx(&(*first, *second))] {
                    Ordering::Less
                } else if rules[pair_to_idx(&(*second, *first))] {
                    Ordering::Greater
                } else {
                    Ordering::Equal
                }
            });

            sorted_update[update.len() / 2] as u32
        }).sum()
}

fn parse(input: &str) -> ([bool; 9999], Vec<Vec<u8>>) {
    let mut rules = [false; 9999];
    let mut updates = Vec::with_capacity(210);

    let mut update_buf = Vec::with_capacity(40);

    let bytes = input.as_bytes();
    let mut state = RuleLeft(0);
    for i in 0..bytes.len() {
        state = match state {
            RuleLeft(left) => {
                match bytes[i] {
                    nb @ b'0'..=b'9' => RuleLeft(left * 10 + (nb - b'0')),
                    b'|' => {
                        RuleRight(left, 0)
                    }
                    _ => unreachable!()
                }
            }
            RuleRight(left, right) => {
                match bytes[i] {
                    nb @ b'0'..=b'9' => RuleRight(left, right * 10 + (nb - b'0')),
                    b'\n' => {
                        rules[pair_to_idx(&(left, right))] = true;
                        RuleEOL
                    }
                    b'\r' => state, // ignore (windows)
                    _ => unreachable!()
                }
            }
            RuleEOL => {
                match bytes[i] {
                    nb @ b'0'..=b'9' => RuleLeft(nb - b'0'),
                    b'\n' => Update(0), // switch mode
                    b'\r' => state, // ignore (windows)
                    _ => unreachable!()
                }
            }
            Update(n) => {
                match bytes[i] {
                    nb @ b'0'..=b'9' => Update(n * 10 + (nb - b'0')),
                    b',' => {
                        update_buf.push(n);
                        Update(0)
                    }
                    b'\n' => {
                        update_buf.push(n);
                        updates.push(update_buf.clone());
                        update_buf.clear();
                        Update(0)
                    }
                    b'\r' => state, // ignore (windows)
                    _ => unreachable!()
                }
            }
        }
    }
    if let Update(n) = state {
        update_buf.push(n);
        updates.push(update_buf);
    } else {
        unreachable!()
    }

    (rules, updates)
}

#[inline]
fn pair_to_idx((first, second): &(u8, u8)) -> usize {
    (*first as usize * 100) + (*second as usize)
}

#[derive(Copy, Clone)]
enum ParsingState {
    RuleLeft(u8),
    RuleRight(u8, u8),
    RuleEOL,
    Update(u8),
}

#[cfg(test)]
mod test {
    use super::*;
    use crate::default_tests;

    const INPUT: &str = "47|53
97|13
97|61
97|47
75|29
61|13
75|53
29|13
97|29
53|29
61|53
97|53
61|29
47|13
75|47
97|75
47|61
75|61
47|29
75|13
53|13

75,47,61,53,29
97,61,53,29,13
75,29,13
75,97,47,61,53
61,13,29
97,13,75,29,47";

    #[test]
    fn text_example_1() {
        assert_eq!(part1(INPUT), 143);
    }

    #[test]
    fn text_example_2() {
        assert_eq!(part2(INPUT), 123);
    }

    default_tests!(day05, 4996, 6311);
}