use crate::day06::GuardDirection::{Down, Left, Right, Up};
use crate::day06::MapEntry::{Empty, Guard, Obstacle, Visited};
use crate::utils::{coords, VecGrid};
use rayon::prelude::*;
use std::cmp::PartialEq;
use std::collections::{HashMap, HashSet};

pub fn part1(input: &str) -> u32 {
    let mut map = parse(input);
    simulate_modify(&mut map.map, map.guard_pos, map.guard_direction);

    map.map
        .iter()
        .filter(|&&entry| match entry {
            Visited => true,
            Guard => true,
            _ => false,
        })
        .count() as u32
}

pub fn part2(input: &str) -> u32 {
    let mut map = parse(input);
    simulate_modify(&mut map.map, map.guard_pos, map.guard_direction);

    let visited_pos: Vec<_> = coords(map.map.height(), map.map.width())
        .filter(|&pos| {
            let entry = *map.map.get(pos);
            entry == Visited || entry == Guard
        })
        .collect();

    visited_pos
        .par_iter()
        .filter(|&&obstacle_pos| {
            let mut curr_map = map.map.clone();
            curr_map.set(obstacle_pos, Obstacle);

            simulate_loop(&curr_map, map.guard_pos, map.guard_direction)
        })
        .count() as u32
}

fn simulate_modify(map: &mut VecGrid<MapEntry>, start: (usize, usize), dir: GuardDirection) {
    let mut guard_pos = start;
    let mut guard_dir = dir;
    let mut guard_offs = guard_dir.movement_offset();

    loop {
        let next_pos = (
            (guard_pos.0 as i32) + guard_offs.0,
            (guard_pos.1 as i32) + guard_offs.1,
        );
        if next_pos.0 < 0
            || next_pos.1 < 0
            || next_pos.0 >= map.height() as i32
            || next_pos.1 >= map.width() as i32
        {
            break;
        }
        let next_pos = to_unsigned(next_pos);

        let front_of_guard = map.get(next_pos);
        match front_of_guard {
            Obstacle => {
                guard_dir = guard_dir.turn();
                guard_offs = guard_dir.movement_offset();
            }
            _ => {
                map.set(next_pos, Guard);
                map.set(guard_pos, Visited);
                guard_pos = next_pos;
            }
        }
    }
}

fn simulate_loop(map: &VecGrid<MapEntry>, start: (usize, usize), dir: GuardDirection) -> bool {
    let mut pos_history: HashMap<(usize, usize), HashSet<GuardDirection>> = HashMap::new();
    let mut guard_pos = start;
    let mut guard_dir = dir;
    let mut guard_offs = guard_dir.movement_offset();

    loop {
        pos_history
            .entry(guard_pos)
            .or_insert_with(HashSet::new)
            .insert(guard_dir);

        let next_pos = (
            (guard_pos.0 as i32) + guard_offs.0,
            (guard_pos.1 as i32) + guard_offs.1,
        );
        if next_pos.0 < 0
            || next_pos.1 < 0
            || next_pos.0 >= map.height() as i32
            || next_pos.1 >= map.width() as i32
        {
            return false;
        }
        let next_pos = to_unsigned(next_pos);
        match pos_history.get(&next_pos) {
            Some(dirs) => {
                if dirs.contains(&guard_dir) {
                    return true;
                }
            }
            None => {}
        }

        let front_of_guard = map.get(next_pos);
        match front_of_guard {
            Obstacle => {
                guard_dir = guard_dir.turn();
                guard_offs = guard_dir.movement_offset();
            }
            _ => {
                guard_pos = next_pos;
            }
        }
    }
}

fn parse(input: &str) -> Map {
    let mut map = vec![];
    let mut guard_pos: Option<(usize, usize)> = None;
    let mut guard_dir: Option<GuardDirection> = None;

    let bytes = input.as_bytes();
    let mut width = 0;
    let mut line = 0;
    let mut line_idx = 0;
    for i in 0..bytes.len() {
        match bytes[i] {
            b'.' => map.push(Empty),
            b'#' => map.push(Obstacle),
            b'^' | b'v' | b'<' | b'>' => {
                guard_pos = Some((line, line_idx));
                guard_dir = match bytes[i] {
                    b'^' => Some(Up),
                    b'v' => Some(Down),
                    b'<' => Some(Left),
                    b'>' => Some(Right),
                    _ => unreachable!(),
                };
                map.push(Empty)
            }
            b'\n' => {
                if width == 0 {
                    width = map.len()
                }

                line += 1;
                line_idx = 0;
                continue; // skip incrementing line_idx
            }
            b'\r' => (),
            _ => unreachable!(),
        }

        line_idx += 1;
    }

    Map {
        map: VecGrid::from(map, width),
        guard_direction: guard_dir.unwrap(),
        guard_pos: guard_pos.unwrap(),
    }
}

struct Map {
    map: VecGrid<MapEntry>,
    guard_pos: (usize, usize),
    guard_direction: GuardDirection,
}

#[derive(Copy, Clone, Debug, PartialEq)]
enum MapEntry {
    Empty,
    Guard,
    Obstacle,
    Visited,
}

#[derive(Copy, Clone, Debug, PartialEq, Eq, Hash)]
enum GuardDirection {
    Up,
    Right,
    Down,
    Left,
}

impl GuardDirection {
    fn movement_offset(&self) -> (i32, i32) {
        match self {
            Up => (-1, 0),
            Right => (0, 1),
            Down => (1, 0),
            Left => (0, -1),
        }
    }

    fn turn(&self) -> GuardDirection {
        match self {
            Up => Right,
            Right => Down,
            Down => Left,
            Left => Up,
        }
    }
}

#[inline]
fn to_unsigned((a, b): (i32, i32)) -> (usize, usize) {
    (a as usize, b as usize)
}

#[cfg(test)]
mod test {
    use super::*;
    use crate::default_tests;

    const INPUT: &str = "....#.....
.........#
..........
..#.......
.......#..
..........
.#..^.....
........#.
#.........
......#...";

    #[test]
    fn test_example_1() {
        assert_eq!(part1(INPUT), 41);
    }

    #[test]
    fn test_example_2() {
        assert_eq!(part2(INPUT), 6);
    }

    default_tests!(day06, 4982, 1663);
}
