use rayon::prelude::*;

pub fn part1(input: &str) -> u64 {
    parse(input)
        .par_iter()
        .filter(|(result, numbers)| {
            let (first, rest) = numbers.split_first().unwrap();
            is_possible_pm(*result, *first, rest)
        })
        .map(|(result, _)| result)
        .sum()
}

pub fn part2(input: &str) -> u64 {
    parse(input)
        .par_iter()
        .filter(|(result, numbers)| {
            let (first, rest) = numbers.split_first().unwrap();
            is_possible_pmc(*result, *first, rest)
        })
        .map(|(result, _)| result)
        .sum()
}

fn parse(input: &str) -> Vec<(u64, Vec<u64>)> {
    input.lines()
        .map(|line| {
            let (result, numbers) = line.split_once(": ").unwrap();

            (
                result.parse().unwrap(),
                numbers.split(' ').map(|x| x.parse().unwrap()).collect()
            )
        }).collect()
}

fn is_possible_pm(target: u64, current: u64, numbers: &[u64]) -> bool {
    if current > target {
        return false;
    }

    if numbers.len() == 0 {
        return target == current;
    }

    let mine = numbers[0];
    let rest = &numbers[1..];
    is_possible_pm(target, current * mine, rest)
        || is_possible_pm(target, current + mine, rest)
}

fn is_possible_pmc(target: u64, current: u64, numbers: &[u64]) -> bool {
    if current > target {
        return false;
    }

    if numbers.len() == 0 {
        return target == current;
    }

    let mine = numbers[0];
    let rest = &numbers[1..];
    is_possible_pmc(target, concat_num(current, mine), rest)
        || is_possible_pmc(target, current * mine, rest)
        || is_possible_pmc(target, current + mine, rest)
}

#[inline]
fn concat_num(a: u64, b: u64) -> u64 {
    (a * 10u64.pow(b.ilog10() + 1)) + b
}

#[cfg(test)]
mod test {
    use crate::default_tests;
    use super::*;

    const INPUT: &str = "190: 10 19
3267: 81 40 27
83: 17 5
156: 15 6
7290: 6 8 6 15
161011: 16 10 13
192: 17 8 14
21037: 9 7 18 13
292: 11 6 16 20
";

    #[test]
    fn test_example_1() {
        assert_eq!(part1(INPUT), 3749)
    }

    #[test]
    fn test_example_2() {
        assert_eq!(part2(INPUT), 11387)
    }

    #[test]
    fn test_ilog10() {
        fn test(n: u64, expected: u64) {
            assert_eq!((n.ilog10() + 1) as u64, expected);
        }

        test(1, 1);
        test(9, 1);
        test(10, 2);
        test(11, 2);
        test(15, 2);
        test(12345, 5);
    }

    #[test]
    fn test_concat_num() {
        fn test(a: u64, b: u64, expected: u64) {
            assert_eq!(concat_num(a, b), expected);
        }

        test(1, 1, 11);
        test(10, 10, 1010);
        test(123, 456, 123456);
        test(15, 6, 156);
    }

    default_tests!(day07, 1620690235709, 145397611075341);
}