use crate::utils::BitField;
use itertools::Itertools;

pub fn part1(input: &str) -> u32 {
    let (positions, rows, cols) = parse(input);

    let mut antinode_field = BitField::new(rows as usize, cols as usize);
    for antenna_pos in positions {
        for positions in antenna_pos.iter().combinations(2) {
            let pos1 = *positions[0];
            let pos2 = *positions[1];

            let [node1, node2] = compute_antinodes(pos1, pos2);
            if check_bounds(node1, rows, cols) {
                antinode_field.set((node1.0 as usize, node1.1 as usize));
            }
            if check_bounds(node2, rows, cols) {
                antinode_field.set((node2.0 as usize, node2.1 as usize));
            }
        }
    }

    antinode_field.count_ones()
}

pub fn part2(input: &str) -> u32 {
    let (positions, rows, cols) = parse(input);

    let mut antinode_field = BitField::new(rows as usize, cols as usize);
    for antenna_pos in positions.iter() {
        for positions in antenna_pos.iter().combinations(2) {
            let pos1 = *positions[0];
            let pos2 = *positions[1];

            antinode_field.set((pos1.0 as usize, pos1.1 as usize));
            antinode_field.set((pos2.0 as usize, pos2.1 as usize));

            let forward_diff = antinode_diff(pos1, pos2);
            let mut next_antinode = antinode_step(pos2, forward_diff);
            while check_bounds(next_antinode, rows, cols) {
                antinode_field.set((next_antinode.0 as usize, next_antinode.1 as usize));
                next_antinode = antinode_step(next_antinode, forward_diff);
            }

            let backwards_diff = (forward_diff.0 * -1, forward_diff.1 * -1);
            let mut next_antinode = antinode_step(pos1, backwards_diff);
            while check_bounds(next_antinode, rows, cols) {
                antinode_field.set((next_antinode.0 as usize, next_antinode.1 as usize));
                next_antinode = antinode_step(next_antinode, backwards_diff);
            }
        }
    }

    antinode_field.count_ones()
}

fn parse(input: &str) -> (Vec<Vec<(i32, i32)>>, i32, i32) {
    let mut freq_lists = vec![Vec::new(); 10 + 26 + 26]; // 0-9 a-z A-Z
    fn remap(c: char) -> usize {
        match c {
            '0'..='9' => c as usize - '0' as usize,
            'a'..='z' => (c as usize - 'a' as usize) + 10,
            'A'..='Z' => (c as usize - 'A' as usize) + 36,
            _ => unreachable!(),
        }
    }

    let mut row = 0;
    let mut col = 0;

    for x in input.lines() {
        col = 0;

        for c in x.chars() {
            if c != '.' {
                freq_lists[remap(c)].push((row, col));
            }

            col += 1;
        }
        row += 1;
    }

    (
        freq_lists.into_iter().filter(|v| !v.is_empty()).collect(),
        row,
        col,
    )
}

#[inline]
fn compute_antinodes(a: (i32, i32), b: (i32, i32)) -> [(i32, i32); 2] {
    let (row_diff, col_diff) = antinode_diff(a, b);

    [
        (b.0 + row_diff, b.1 + col_diff),
        (a.0 - row_diff, a.1 - col_diff),
    ]
}

#[inline]
fn antinode_step(a: (i32, i32), (row_diff, col_diff): (i32, i32)) -> (i32, i32) {
    (a.0 + row_diff, a.1 + col_diff)
}

#[inline]
fn antinode_diff(a: (i32, i32), b: (i32, i32)) -> (i32, i32) {
    (b.0 - a.0, b.1 - a.1)
}

#[inline]
fn check_bounds(pos: (i32, i32), rows: i32, cols: i32) -> bool {
    pos.0 >= 0 && pos.1 >= 0 && pos.0 < rows && pos.1 < cols
}

#[cfg(test)]
mod test {
    use super::*;
    use crate::default_tests;

    const INPUT: &str = "............
........0...
.....0......
.......0....
....0.......
......A.....
............
............
........A...
.........A..
............
............";

    #[test]
    fn test_example_1() {
        assert_eq!(part1(INPUT), 14)
    }

    #[test]
    fn test_bounds_1() {
        assert_eq!(part1(".0\n0."), 0)
    }
    #[test]
    fn test_bounds_2() {
        assert_eq!(part1("..0\n.0.\n..."), 1)
    }

    #[test]
    fn test_bounds_3() {
        assert_eq!(part1("...\n.0.\n0.."), 1)
    }

    #[test]
    fn test_overlap_1() {
        assert_eq!(part1("..0\n.00\n..0"), 5)
    }

    #[test]
    fn test_example_2() {
        assert_eq!(part2(INPUT), 34)
    }

    default_tests!(day08, 400, 1280);
}
