use std::cmp::Ordering;
use std::fmt::Debug;

const ASCII_ZERO: u8 = b'0';

pub fn part1(input: &str) -> u64 {
    let parsed = input.as_bytes();

    let mut checksum = 0u64;

    let mut left_head = 0usize;
    let mut left_id = 0usize;
    let mut left_pos = 0usize;

    let mut right_head = match parsed.len() % 2 {
        // this is space
        0 => parsed.len() - 2,
        // this is a file
        1 => parsed.len() - 1,
        _ => unreachable!(),
    };
    // fsfsfsf = l=7 / idx = 6
    //       ^
    // 0 1 2 3
    let mut right_id = right_head / 2;
    let mut right_entries_left = (parsed[right_head] - ASCII_ZERO) as usize;

    'outer: loop {
        if left_head >= right_head {
            break;
        }

        // starts with file, which we can already accumulate
        // pos      5   6   7   8   9
        // file id  2   2   2   2   2
        //
        //   (2*5) + (2*6) + ...
        // = 2 * (5 + 6 + ...)
        // S=(β−α+1)(α+β)*1/2
        // 2 * (9-5+1)(5+9)/2 = 2 * (5*14/2) = 2 * 30 = 60
        let file_length = (parsed[left_head] - ASCII_ZERO) as usize;
        checksum += left_id as u64 * range_sum(left_pos, left_pos + file_length);
        left_pos += file_length;

        // now the empty space
        left_head += 1;
        let mut space_length = (parsed[left_head] - ASCII_ZERO) as usize;
        while space_length > 0 {
            // grab some files from the right
            let take = match space_length.cmp(&right_entries_left) {
                Ordering::Less => {
                    // we have less space than there are right entries available, take some of them
                    space_length
                }
                Ordering::Greater | Ordering::Equal => {
                    // we have more or equal space than there are right entries available, so take them all
                    right_entries_left
                }
            };

            checksum += right_id as u64 * range_sum(left_pos, left_pos + take);
            left_pos += take;
            space_length -= take;
            right_entries_left -= take;

            // next one please
            if right_entries_left == 0 {
                if right_head < 2 || right_head - 2 <= left_head {
                    // trying to exit left, edge casey
                    break 'outer;
                }
                right_head -= 2;
                right_id -= 1;
                right_entries_left = (parsed[right_head] - ASCII_ZERO) as usize;
            }
        }

        // file for the next iteration
        left_head += 1;
        left_id += 1;
    }

    if right_entries_left != 0 {
        checksum += right_id as u64 * range_sum(left_pos, left_pos + right_entries_left);
    }

    checksum
}

pub fn part2(input: &str) -> u64 {
    let bytes = input.as_bytes();
    let half_input_length = bytes.len() / 2;

    let mut spaces = Vec::with_capacity(half_input_length);
    let mut files = Vec::with_capacity(half_input_length);

    let mut id = 0;
    let mut offset = 0;
    for i in 0..bytes.len() {
        let block = bytes[i];
        let length = block - ASCII_ZERO;
        if i % 2 == 0 {
            files.push(File { offset, id, length });

            id += 1;
        } else {
            spaces.push(Space { offset, length });
        }
        offset += length as u32;
    }

    let mut checksum = 0u64;
    let mut cache: [usize; 10] = [0; 10];

    for file in files.iter().rev() {
        let mut found = false;
        for i in cache[file.length as usize]..(file.id as usize) {
            let space = &mut spaces[i];
            if space.length >= file.length {
                let offset = space.offset as usize;
                checksum += file.id as u64 * range_sum(offset, offset + file.length as usize);
                space.length -= file.length;
                space.offset += file.length as u32;
                cache[file.length as usize] = i;
                found = true;
                break;
            }
        }

        if !found {
            let offset_u = file.offset as usize;
            checksum += file.id as u64 * range_sum(offset_u, offset_u + file.length as usize);
        }
    }

    checksum
}

#[inline]
fn range_sum(start: usize, end_exclusive: usize) -> u64 {
    if start >= end_exclusive {
        return 0;
    }

    let lower = start as u64;
    // SAFETY: Start and end_exclusive are both unsigned, so at least zero.
    // End cannot be zero, causing an underflow, because end must be > start to end up here.
    let upper = unsafe { end_exclusive.unchecked_sub(1) } as u64;

    (upper - lower + 1) * (lower + upper) / 2
}

#[derive(Debug, Clone)]
struct File {
    offset: u32,
    id: u16,
    length: u8,
}

#[derive(Debug, Clone)]
struct Space {
    offset: u32,
    length: u8,
}

#[cfg(test)]
mod test {
    use super::*;
    use crate::default_tests;

    const INPUT: &str = "2333133121414131402";

    #[test]
    fn test_example_1() {
        assert_eq!(part1(INPUT), 1928)
    }

    #[test]
    fn test_not_enough() {
        assert_eq!(part1("054"), 6);
    }

    #[test]
    fn test_too_many() {
        assert_eq!(part1("04015"), 20);
    }

    #[test]
    fn test_fits_just_right() {
        assert_eq!(part1("504"), 26);
    }
    #[test]
    fn test_even() {
        assert_eq!(part1("1234"), 6);
    }
    #[test]
    fn test_one_in_each_with_rest() {
        assert_eq!(part1("111111115"), 124);
    }

    #[test]
    fn test_hole_with_space_at_end() {
        assert_eq!(part1("14103"), 16);
    }

    #[test]
    fn test_read_head_passing_write_head() {
        assert_eq!(part1("12113"), 17);
    }

    #[test]
    fn test_example_2() {
        assert_eq!(part2(INPUT), 2858)
    }

    default_tests!(day09, 6_283_404_590_840, 6_304_576_012_713);
}
