use crate::utils::{coords, BitField};

const ASCII_ZERO: u8 = b'0';

pub fn part1(input: &str) -> u32 {
    let parsed = parse(input);

    coords(parsed.len(), parsed[0].len())
        .filter(|&(row, col)| parsed[row][col] == 0)
        .map(|coord| trailhead_score(coord, &parsed))
        .sum()
}

pub fn part2(input: &str) -> u32 {
    let parsed = parse(input);

    coords(parsed.len(), parsed[0].len())
        .filter(|&(row, col)| parsed[row][col] == 0)
        .map(|coord| trailhead_rating(coord, &parsed))
        .sum()
}

#[inline]
fn trailhead_score(coord: (usize, usize), grid: &Vec<Vec<u8>>) -> u32 {
    let visited = BitField::new(grid.len(), grid[0].len());
    let mut strategy = DistinctTopStrategy { visited };

    follow_incline(&mut strategy, 0, coord, grid)
}

#[inline]
fn trailhead_rating(coord: (usize, usize), grid: &Vec<Vec<u8>>) -> u32 {
    follow_incline(&mut DistinctPathsStrategy, 0, coord, grid)
}

#[inline]
fn follow_incline<S: DistinctStrategy>(
    strategy: &mut S,
    last_value: u8,
    coord: (usize, usize),
    grid: &Vec<Vec<u8>>,
) -> u32 {
    if let Some(ret) = strategy.should_return(coord, last_value) {
        return ret;
    }

    let next_value = last_value + 1;

    [
        apply_offset(coord, (-1, 0), grid),
        apply_offset(coord, (1, 0), grid),
        apply_offset(coord, (0, 1), grid),
        apply_offset(coord, (0, -1), grid),
    ]
    .iter()
    .filter_map(|new_coord| match new_coord {
        Some(new_coord) if grid[new_coord.0][new_coord.1] == next_value => Some(new_coord),
        _ => None,
    })
    .map(|&new_coord| follow_incline(strategy, next_value, new_coord, grid))
    .sum()
}

#[inline]
fn apply_offset(
    (row, col): (usize, usize),
    (row_offset, col_offset): (i8, i8),
    grid: &Vec<Vec<u8>>,
) -> Option<(usize, usize)> {
    let new_row = row.checked_add_signed(row_offset as isize)?;
    let new_col = col.checked_add_signed(col_offset as isize)?;

    (new_row < grid.len() && new_col < grid[0].len()).then_some((new_row, new_col))
}

fn parse(input: &str) -> Vec<Vec<u8>> {
    input
        .lines()
        .map(|line| line.as_bytes().iter().map(|b| b - ASCII_ZERO).collect())
        .collect()
}

trait DistinctStrategy {
    fn should_return(&mut self, coord: (usize, usize), last_value: u8) -> Option<u32>;
}

struct DistinctTopStrategy {
    visited: BitField,
}

impl<'a> DistinctStrategy for DistinctTopStrategy {
    #[inline(always)]
    fn should_return(&mut self, coord: (usize, usize), last_value: u8) -> Option<u32> {
        if self.visited.get(coord) {
            return Some(0);
        }
        self.visited.set(coord);

        if last_value == 9 {
            Some(1)
        } else {
            None
        }
    }
}

struct DistinctPathsStrategy;

impl DistinctStrategy for DistinctPathsStrategy {
    #[inline(always)]
    fn should_return(&mut self, _coord: (usize, usize), last_value: u8) -> Option<u32> {
        if last_value == 9 {
            Some(1)
        } else {
            None
        }
    }
}

#[cfg(test)]
mod test {
    use super::*;
    use crate::default_tests;

    const INPUT: &str = "89010123
78121874
87430965
96549874
45678903
32019012
01329801
10456732";

    #[test]
    fn test_example_1() {
        assert_eq!(part1(INPUT), 36);
    }

    #[test]
    fn test_example_2() {
        assert_eq!(part2(INPUT), 81);
    }

    default_tests!(day10, 535, 1186);
}
