#[cfg(windows)]
pub(crate) const LINE_SEPARATOR: &str = "\r\n";
#[cfg(windows)]
pub(crate) const LINE_SEPARATOR_LEN: usize = LINE_SEPARATOR.len();

#[cfg(unix)]
pub(crate) const LINE_SEPARATOR: &str = "\n";
#[cfg(unix)]
pub(crate) const LINE_SEPARATOR_LEN: usize = LINE_SEPARATOR.len();

#[inline(always)]
pub fn coords(lines: usize, rows: usize) -> impl Iterator<Item = (usize, usize)> {
    (0..lines)
        .into_iter()
        .flat_map(move |line| (0..rows).into_iter().map(move |row| (line, row)))
}

pub struct BitField {
    data: Vec<u64>,
    width: usize,
}

impl BitField {
    pub fn new(height: usize, width: usize) -> Self {
        let size = (height * width + 63) / 64;
        BitField {
            data: vec![0; size],
            width,
        }
    }

    #[inline(always)]
    pub fn get(&self, (row, col): (usize, usize)) -> bool {
        let index = row * self.width + col;
        let (word_index, bit_index) = (index / 64, index % 64);
        (self.data[word_index] & (1 << bit_index)) != 0
    }

    #[inline(always)]
    pub fn set(&mut self, (row, col): (usize, usize)) {
        let index = row * self.width + col;
        let (word_index, bit_index) = (index / 64, index % 64);
        self.data[word_index] |= 1 << bit_index;
    }

    #[inline(always)]
    pub fn count_ones(&self) -> u32 {
        self.data.iter().map(|bits| bits.count_ones()).sum()
    }
}

pub struct VecGrid<T> {
    backing: Vec<T>,
    width: usize,
}

impl<T> VecGrid<T> {
    pub fn from(vec: Vec<T>, width: usize) -> Self {
        let grid = VecGrid {
            backing: vec,
            width,
        };

        if grid.backing.len() < grid.width() * grid.height() {
            panic!("backing array does not fill the whole rectangle")
        }

        grid
    }

    #[inline(always)]
    fn index(&self, (row, col): (usize, usize)) -> usize {
        row * self.width + col
    }

    #[inline(always)]
    pub fn get(&self, pos: (usize, usize)) -> &T {
        &self.backing[self.index(pos)]
    }

    #[inline(always)]
    pub fn get_mut(&mut self, pos: (usize, usize)) -> &mut T {
        let idx = self.index(pos);
        &mut self.backing[idx]
    }

    #[inline(always)]
    pub fn set(&mut self, pos: (usize, usize), value: T) {
        let idx = self.index(pos);
        self.backing[idx] = value;
    }

    #[inline(always)]
    pub fn width(&self) -> usize {
        self.width
    }

    #[inline(always)]
    pub fn height(&self) -> usize {
        (self.backing.len() + self.width - 1) / self.width
    }

    #[inline(always)]
    pub fn len(&self) -> usize {
        self.height() * self.width()
    }

    pub fn iter(&self) -> impl Iterator<Item = &T> {
        self.backing.iter()
    }
}

impl<T: Clone> Clone for VecGrid<T> {
    fn clone(&self) -> Self {
        let new_backing = self.backing.clone();

        VecGrid {
            backing: new_backing,
            width: self.width,
        }
    }
}

#[cfg(test)]
mod vec_grid_test {
    use crate::utils::VecGrid;
    use std::panic::catch_unwind;

    fn standard_grid() -> VecGrid<u8> {
        let vec = vec![1, 2, 3, 4, 5, 6];
        VecGrid::from(vec, 3)
    }

    #[test]
    fn creates() {
        let grid = standard_grid();

        assert_eq!(grid.width(), 3);
        assert_eq!(grid.height(), 2);
        assert_eq!(grid.len(), 6);
    }

    #[test]
    fn rejects_invalid_backing_arrays() {
        let backing = vec![1, 2, 3, 4, 5];

        assert_eq!(catch_unwind(|| VecGrid::from(backing, 2)).is_err(), true);
    }

    #[test]
    fn gets() {
        let grid = standard_grid();

        assert_eq!(grid.get((0, 0)), &1);
        assert_eq!(grid.get((0, 1)), &2);
        assert_eq!(grid.get((1, 0)), &4);
        assert_eq!(grid.get((1, 2)), &6);
    }

    #[test]
    fn sets() {
        let mut grid = standard_grid();

        grid.set((0, 0), 99);
        assert_eq!(grid.get((0, 0)), &99);
        assert_eq!(grid.get((0, 1)), &2);
        assert_eq!(grid.get((0, 2)), &3);
        assert_eq!(grid.get((1, 0)), &4);
        assert_eq!(grid.get((1, 1)), &5);
        assert_eq!(grid.get((1, 2)), &6);

        grid.set((0, 2), 99);
        assert_eq!(grid.get((0, 0)), &99);
        assert_eq!(grid.get((0, 1)), &2);
        assert_eq!(grid.get((0, 2)), &99);
        assert_eq!(grid.get((1, 0)), &4);
        assert_eq!(grid.get((1, 1)), &5);
        assert_eq!(grid.get((1, 2)), &6);

        grid.set((1, 0), 99);
        assert_eq!(grid.get((0, 0)), &99);
        assert_eq!(grid.get((0, 1)), &2);
        assert_eq!(grid.get((0, 2)), &99);
        assert_eq!(grid.get((1, 0)), &99);
        assert_eq!(grid.get((1, 1)), &5);
        assert_eq!(grid.get((1, 2)), &6);
    }

    #[test]
    fn iterates() {
        let grid = standard_grid();
        let mut iter = grid.iter();

        assert_eq!(iter.next(), Some(&1));
        assert_eq!(iter.next(), Some(&2));
        assert_eq!(iter.next(), Some(&3));
        assert_eq!(iter.next(), Some(&4));
        assert_eq!(iter.next(), Some(&5));
        assert_eq!(iter.next(), Some(&6));
        assert_eq!(iter.next(), None);
    }
}
